@@ -1,5 +1,5 @@
 # ----------------------------------------------------------------------------
-# Copyright (c) 2016-2022, QIIME 2 development team.
+# Copyright (c) 2016-2023, QIIME 2 development team.
 #
 # Distributed under the terms of the Modified BSD License.
 #
@@ -7,6 +7,7 @@
 # ----------------------------------------------------------------------------
 
 import os
+from typing import Literal
 
 import click
 
@@ -80,41 +81,130 @@ def export_data(input_path, output_path, output_format):
     click.echo(CONFIG.cfg_style('success', success))
 
 
-def show_importable_types(ctx, param, value):
-    if not value or ctx.resilient_parsing:
-        return
-
-    import q2cli.util
-
-    importable_types = sorted(q2cli.util.get_plugin_manager().importable_types,
-                              key=repr)
-
-    if importable_types:
-        for name in importable_types:
-            click.echo(name)
+def _print_descriptions(descriptions, tsv):
+    if tsv:
+        for value, description in descriptions.items():
+            click.echo(f"{value}\t", nl=False)
+            if description:
+                click.echo(_deformat_description(description))
+            else:
+                click.echo()
     else:
-        click.echo('There are no importable types in the current deployment.')
-
-    ctx.exit()
+        import textwrap
+        tabsize = 8
+        for value, description in descriptions.items():
+            click.secho(value, bold=True)
+            if description:
+                description = _deformat_description(description)
+                wrapped_description = textwrap.wrap(description,
+                                                    width=72-tabsize,
+                                                    initial_indent='\t',
+                                                    subsequent_indent='\t',
+                                                    tabsize=tabsize)
+                for line in wrapped_description:
+                    click.echo(f"{line}")
+            else:
+                click.secho("\tNo description", italic=True)
+            click.echo()
 
 
-def show_importable_formats(ctx, param, value):
-    if not value or ctx.resilient_parsing:
-        return
+def _deformat_description(description):
+    import re
+    deformatted = re.sub(r"[\t\n]+", ' ', description)
+    despaced = re.sub(r" +", ' ', deformatted)
+    return despaced
 
-    import q2cli.util
 
-    importable_formats = sorted(
-            q2cli.util.get_plugin_manager().importable_formats)
-
-    if importable_formats:
-        for format in importable_formats:
-            click.echo(format)
+def _get_matches(words, possibilities, strict=False):
+    from difflib import get_close_matches
+    if strict:
+        cutoff = 1
     else:
-        click.echo('There are no importable formats '
-                   'in the current deployment.')
+        cutoff = 0.6
 
-    ctx.exit()
+    matches = set()
+    num_possibilities = len(possibilities)
+    for word in words:
+        matches.update(get_close_matches(word,
+                                         possibilities,
+                                         n=num_possibilities,
+                                         cutoff=cutoff))
+        # substring search
+        if cutoff != 1:
+            for possibility in possibilities:
+                if word.lower() in possibility.lower():
+                    matches.add(possibility)
+
+    return list(matches)
+
+
+@tools.command(
+        name='list-types',
+        help='List the available semantic types.',
+        short_help='',
+        cls=ToolCommand
+)
+@click.argument('queries', nargs=-1)
+@click.option('--strict', is_flag=True,
+              help='Show only exact matches for the type argument(s).')
+@click.option('--tsv', is_flag=True,
+              help='Print as machine readable tab-separated values.')
+def show_types(queries, strict, tsv):
+    pm = q2cli.util.get_plugin_manager()
+
+    if len(queries) > 0:
+        matches = _get_matches(queries, list(pm.artifact_classes), strict)
+    else:
+        matches = sorted(list(pm.artifact_classes))
+
+    descriptions = {}
+    for match in matches:
+        description = pm.artifact_classes[match].description
+        descriptions[match] = description
+
+    _print_descriptions(descriptions, tsv)
+
+
+@tools.command(
+        name='list-formats',
+        help='List the available formats.',
+        short_help='',
+        cls=ToolCommand
+)
+@click.argument('queries', nargs=-1)
+@click.option('--importable', is_flag=True,
+              help='List the importable formats.')
+@click.option('--exportable', is_flag=True,
+              help='List the exportable formats.')
+@click.option('--strict', is_flag=True,
+              help='Show only exact matches for the format argument(s).')
+@click.option('--tsv', is_flag=True,
+              help='Print as machine readable tab-separated values.')
+def show_formats(queries, importable, exportable, strict, tsv):
+    if importable and exportable:
+        raise click.UsageError("'--importable' and '--exportable' flags are "
+                               "mutually exclusive.")
+    if not importable and not exportable:
+        raise click.UsageError("One of '--importable' or '--exportable' flags "
+                               "is required.")
+
+    pm = q2cli.util.get_plugin_manager()
+    portable_formats = pm.importable_formats if importable \
+        else pm.exportable_formats
+
+    if len(queries) > 0:
+        matches = _get_matches(queries, portable_formats.keys(), strict)
+    else:
+        matches = sorted(portable_formats.keys())
+
+    descriptions = {}
+    for match in matches:
+        docstring = portable_formats[match].format.__doc__
+        first_docstring_line = docstring.split('\n\n')[0].strip() \
+            if docstring else ''
+        descriptions[match] = first_docstring_line
+
+    _print_descriptions(descriptions, tsv)
 
 
 @tools.command(name='import',
@@ -126,7 +216,7 @@ def show_importable_formats(ctx, param, value):
                     cls=ToolCommand)
 @click.option('--type', required=True,
               help='The semantic type of the artifact that will be created '
-                   'upon importing. Use --show-importable-types to see what '
+                   'upon importing. Use `qiime tools list-types` to see what '
                    'importable semantic types are available in the current '
                    'deployment.')
 @click.option('--input-path', required=True,
@@ -140,29 +230,22 @@ def show_importable_formats(ctx, param, value):
 @click.option('--input-format', required=False,
               help='The format of the data to be imported. If not provided, '
                    'data must be in the format expected by the semantic type '
-                   'provided via --type.')
-@click.option('--show-importable-types', is_flag=True, is_eager=True,
-              callback=show_importable_types, expose_value=False,
-              help='Show the semantic types that can be supplied to --type '
-                   'to import data into an artifact.')
-@click.option('--show-importable-formats', is_flag=True, is_eager=True,
-              callback=show_importable_formats, expose_value=False,
-              help='Show formats that can be supplied to --input-format to '
-                   'import data into an artifact.')
-def import_data(type, input_path, output_path, input_format):
-    import qiime2.sdk
-    import qiime2.plugin
+                   'provided via --type. Use `qiime tools list-formats '
+                   '--importable` to see which formats of input data are '
+                   'importable.')
+@click.option('--validate-level', default='max',
+              type=click.Choice(['min', 'max']),
+              help='How much to validate the imported data before creating the'
+                   ' artifact. A value of "max" will generally read the entire'
+                   ' file or directory, whereas "min" will not usually do so.'
+                   ' [default: "max"]')
+def import_data(type, input_path, output_path, input_format, validate_level):
+
     from q2cli.core.config import CONFIG
-    try:
-        artifact = qiime2.sdk.Artifact.import_data(type, input_path,
-                                                   view_type=input_format)
-    except qiime2.plugin.ValidationError as e:
-        header = 'There was a problem importing %s:' % input_path
-        q2cli.util.exit_with_error(e, header=header, traceback=None)
-    except Exception as e:
-        header = 'An unexpected error has occurred:'
-        q2cli.util.exit_with_error(e, header=header)
+
+    artifact = _import(type, input_path, input_format, validate_level)
     artifact.save(output_path)
+
     if input_format is None:
         input_format = artifact.format.__name__
 
@@ -181,7 +264,7 @@ def import_data(type, input_path, output_path, input_format):
                 type=click.Path(exists=True, file_okay=True, dir_okay=False,
                                 readable=True), metavar=_COMBO_METAVAR)
 @click.option('--tsv/--no-tsv', default=False,
-              help='Print as machine-readable tab separated values.')
+              help='Print as machine-readable tab-separated values.')
 def peek(paths, tsv):
     import qiime2.sdk
     from q2cli.core.config import CONFIG
@@ -454,7 +537,7 @@ def view(visualization_path, index_extension):
     if index_extension.startswith('.'):
         index_extension = index_extension[1:]
 
-    visualization = _load_input(visualization_path, view=True)[0]
+    _, visualization = _load_input(visualization_path, view=True)[0]
     if not isinstance(visualization, Visualization):
         raise click.BadParameter(
             '%s is not a QIIME 2 Visualization. Only QIIME 2 Visualizations '
@@ -710,6 +793,75 @@ def cache_store(cache, artifact_path, key):
     click.echo(CONFIG.cfg_style('success', success))
 
 
+@tools.command(name='cache-import',
+               short_help='Imports data into an Artifact in the cache under a '
+                          'key.',
+               help='Imports data into an Artifact in the cache under a key.',
+               cls=ToolCommand)
+@click.option('--type', required=True,
+              help='The semantic type of the artifact that will be created '
+                   'upon importing. Use `qiime tools list-types` to see what '
+                   'importable semantic types are available in the current '
+                   'deployment.')
+@click.option('--input-path', required=True,
+              type=click.Path(exists=True, file_okay=True, dir_okay=True,
+                              readable=True),
+              help='Path to file or directory that should be imported.')
+@click.option('--cache', required=True,
+              type=click.Path(exists=True, file_okay=False, dir_okay=True,
+                              readable=True),
+              help='Path to an existing cache to save into.')
+@click.option('--key', required=True,
+              help='The key to save the artifact under (must be a valid '
+                   'Python identifier).')
+@click.option('--input-format', required=False,
+              help='The format of the data to be imported. If not provided, '
+                   'data must be in the format expected by the semantic type '
+                   'provided via --type. Use `qiime tools list-formats '
+                   '--importable` to see which formats of input data are '
+                   'importable.')
+@click.option('--validate-level', required=False, default='max',
+              type=click.Choice(['min', 'max']),
+              help='How much to validate the imported data before creating the'
+                   ' artifact. A value of "max" will generally read the entire'
+                   ' file or directory, whereas "min" will not usually do so.'
+                   ' [default: "max"]')
+def cache_import(type, input_path, cache, key, input_format, validate_level):
+    from qiime2 import Cache
+    from q2cli.core.config import CONFIG
+
+    artifact = _import(type, input_path, input_format, validate_level)
+    _cache = Cache(cache)
+    _cache.save(artifact, key)
+
+    if input_format is None:
+        input_format = artifact.format.__name__
+
+    success = 'Imported %s as %s to %s:%s' % (input_path,
+                                              input_format,
+                                              cache,
+                                              key)
+    click.echo(CONFIG.cfg_style('success', success))
+
+
+def _import(type, input_path, input_format, validate_level):
+    import qiime2.sdk
+    import qiime2.plugin
+
+    try:
+        artifact = qiime2.sdk.Artifact.import_data(
+            type, input_path, view_type=input_format,
+            validate_level=validate_level)
+    except qiime2.plugin.ValidationError as e:
+        header = 'There was a problem importing %s:' % input_path
+        q2cli.util.exit_with_error(e, header=header, traceback=None)
+    except Exception as e:
+        header = 'An unexpected error has occurred:'
+        q2cli.util.exit_with_error(e, header=header)
+
+    return artifact
+
+
 @tools.command(name='cache-fetch',
                short_help='Fetches an artifact out of a cache into a .qza.',
                help='Fetches the artifact saved to the specified cache under '
@@ -767,11 +919,13 @@ def cache_status(cache):
             for key in _cache.get_keys():
                 key_values = _cache.read_key(key)
 
-                if (data := key_values['data']) is not None:
+                if 'data' in key_values:
+                    data = key_values['data']
                     data_output.append(
                         'data: %s -> %s' %
                         (key, str(Result.peek(_cache.data / data))))
-                elif (pool := key_values['pool']) is not None:
+                elif 'pool' in key_values:
+                    pool = key_values['pool']
                     pool_output.append(
                         'pool: %s -> size = %s' %
                         (key, str(len(os.listdir(_cache.pools / pool)))))
@@ -796,3 +950,273 @@ def cache_status(cache):
     success = "Status of the cache at the path '%s':\n\n%s" % \
         (cache, output)
     click.echo(CONFIG.cfg_style('success', success))
+
+
+replay_in_fp_help = (
+    'filepath to a QIIME 2 Archive (.qza or .qzv) or directory of Archives'
+)
+replay_recurse_help = (
+    'if in-fp is a directory, will also search sub-directories when finding '
+    'Archives to parse'
+)
+replay_validate_checksums_help = (
+    'check that replayed archives are intact and uncorrupted'
+)
+replay_parse_metadata_help = (
+    'parse the original metadata captured in provenance for review or replay'
+)
+replay_use_recorded_metadata_help = (
+    're-use the original metadata captured in provenance'
+)
+replay_suppress_header_help = (
+   'do not write header/footer blocks in the output script'
+)
+replay_verbose_help = (
+    'print status messages to stdout while processing'
+)
+replay_dump_recorded_metadata_help = (
+    'write the original metadata captured in provenance to disk in the '
+    '--metadata-out-dir directory'
+)
+
+
+@tools.command(name='replay-provenance', cls=ToolCommand)
+@click.option('--in-fp', required=True, help=replay_in_fp_help)
+@click.option('--recurse/--no-recurse',
+              default=False,
+              show_default=True,
+              help=replay_recurse_help)
+@click.option('--usage-driver',
+              default='cli',
+              show_default=True,
+              help='the target interface for your replay script',
+              type=click.Choice(['python3', 'cli'], case_sensitive=False))
+@click.option('--validate-checksums/--no-validate-checksums',
+              default=True,
+              show_default=True,
+              help=replay_validate_checksums_help)
+@click.option('--parse-metadata/--no-parse-metadata',
+              default=True,
+              show_default=True,
+              help=replay_parse_metadata_help)
+@click.option('--use-recorded-metadata/--no-use-recorded-metadata',
+              default=False,
+              show_default=True,
+              help=replay_use_recorded_metadata_help)
+@click.option('--suppress-header/--no-suppress-header',
+              default=False,
+              show_default=True,
+              help=replay_suppress_header_help)
+@click.option('--verbose/--no-verbose',
+              default=True,
+              show_default=True,
+              help=replay_verbose_help)
+@click.option('--dump-recorded-metadata/--no-dump-recorded-metadata',
+              default=True,
+              show_default=True,
+              help=replay_dump_recorded_metadata_help)
+@click.option('--metadata-out-dir',
+              default='',
+              show_default=True,
+              help=('the directory where captured study metadata '
+                    'should be written if --dump-recorded-metadata. This '
+                    'often produces many outputs, so a dedicated directory '
+                    'should generally be used. Creates the directory if it '
+                    'does not already exist. By default, metadata is written '
+                    'to `${PWD}/recorded_metadata/`'))
+@click.option('--out-fp',
+              required=True,
+              type=click.Path(exists=False, writable=True),
+              help='the filepath where your replay script should be written')
+def provenance_replay(
+    in_fp: str,
+    out_fp: str,
+    usage_driver: Literal['python3', 'cli'],
+    recurse: bool = False,
+    validate_checksums: bool = True,
+    parse_metadata: bool = True,
+    use_recorded_metadata: bool = False,
+    suppress_header: bool = False,
+    verbose: bool = True,
+    dump_recorded_metadata: bool = True,
+    metadata_out_dir: str = ''
+):
+    """
+    Replay provenance from a QIIME 2 Artifact filepath to a written executable
+    """
+    from qiime2.core.archive.provenance_lib.replay import replay_provenance
+    from qiime2.sdk.util import get_available_usage_drivers
+
+    usage_drivers = get_available_usage_drivers()
+    try:
+        usage_driver_type = usage_drivers[usage_driver]
+    except KeyError:
+        msg = (
+            f'The {usage_driver} usage driver is not available in the '
+            'current environment.'
+        )
+        raise ValueError(msg)
+
+    replay_provenance(
+        usage_driver=usage_driver_type,
+        payload=in_fp,
+        out_fp=out_fp,
+        validate_checksums=validate_checksums,
+        parse_metadata=parse_metadata,
+        recurse=recurse,
+        use_recorded_metadata=use_recorded_metadata,
+        suppress_header=suppress_header,
+        verbose=verbose,
+        dump_recorded_metadata=dump_recorded_metadata,
+        md_out_dir=metadata_out_dir
+    )
+    filename = os.path.realpath(out_fp)
+    click.echo(f'{usage_driver} replay script written to {filename}')
+
+
+@tools.command(name='replay-citations', cls=ToolCommand)
+@click.option('--in-fp', required=True, help=replay_in_fp_help)
+@click.option('--recurse/--no-recurse',
+              default=False,
+              show_default=True,
+              help=replay_recurse_help)
+@click.option('--deduplicate/--no-deduplicate',
+              default=True,
+              show_default=True,
+              help=('If deduplicate, duplicate citations will be removed '
+                    'heuristically, e.g. by comparing DOI fields. '
+                    'This greatly reduces manual curation of reference lists, '
+                    'but introduces a small risk of reference loss.'))
+@click.option('--suppress-header/--no-suppress-header',
+              default=False,
+              show_default=True,
+              help=replay_suppress_header_help)
+@click.option('--verbose/--no-verbose',
+              default=True,
+              show_default=True,
+              help=replay_verbose_help)
+@click.option('--out-fp',
+              required=True,
+              type=click.Path(exists=False, writable=True),
+              help='the filepath where your bibtex file should be written')
+def citations_replay(
+    in_fp: str,
+    out_fp: str,
+    recurse: bool = False,
+    deduplicate: bool = True,
+    suppress_header: bool = False,
+    verbose: bool = True
+):
+    """
+    Reports all citations from a QIIME 2 Artifact or directory of Artifacts,
+    with the goal of improving and simplifying attribution of/in published
+    work.
+
+    Not for use in reporting e.g. software versions used in an analysis, as
+    deduplication removes duplicate references with different plugin versions.
+    """
+    from qiime2.core.archive.provenance_lib.parse import ProvDAG
+    from qiime2.core.archive.provenance_lib.replay import replay_citations
+
+    dag = ProvDAG(in_fp, verbose=verbose, recurse=recurse)
+    replay_citations(
+        dag,
+        out_fp=out_fp,
+        deduplicate=deduplicate,
+        suppress_header=suppress_header
+    )
+    filename = os.path.realpath(out_fp)
+    click.echo(f'citations bibtex file written to {filename}')
+
+
+@tools.command(name='replay-supplement', cls=ToolCommand)
+@click.option('--in-fp', required=True,
+              help='filepath to a QIIME 2 Archive or directory of Archives')
+@click.option('--recurse/--no-recurse',
+              default=False,
+              show_default=True,
+              help=('if in-fp is a directory, will also search sub-directories'
+                    ' when finding .qza/.qzv files to parse'))
+@click.option('--deduplicate/--no-deduplicate',
+              default=True,
+              show_default=True,
+              help=('If deduplicate, duplicate citations will be removed '
+                    'heuristically, e.g. by comparing DOI fields. '
+                    'This greatly reduces manual curation of reference lists, '
+                    'but introduces a small risk of reference loss.'))
+@click.option('--validate-checksums/--no-validate-checksums',
+              default=True,
+              show_default=True,
+              help=replay_validate_checksums_help)
+@click.option('--parse-metadata/--no-parse-metadata',
+              default=True,
+              show_default=True,
+              help=replay_parse_metadata_help)
+@click.option('--use-recorded-metadata/--no-use-recorded-metadata',
+              default=False,
+              show_default=True,
+              help=replay_use_recorded_metadata_help)
+@click.option('--suppress-header/--no-suppress-header',
+              default=False,
+              show_default=True,
+              help=replay_suppress_header_help)
+@click.option('--verbose/--no-verbose',
+              default=True,
+              show_default=True,
+              help=replay_verbose_help)
+@click.option('--dump-recorded-metadata/--no-dump-recorded-metadata',
+              default=True,
+              show_default=True,
+              help='write the original metadata captured in provenance to '
+                   'recorded_metadata/ inside the archive')
+@click.option('--out-fp',
+              required=True,
+              type=click.Path(exists=False, writable=True),
+              help='the filepath where your reproduciblity supplement zipfile '
+                   'should be written')
+def supplement_replay(
+    in_fp: str,
+    out_fp: str,
+    validate_checksums: bool = True,
+    parse_metadata: bool = True,
+    use_recorded_metadata: bool = False,
+    recurse: bool = False,
+    deduplicate: bool = True,
+    suppress_header: bool = False,
+    verbose: bool = True,
+    dump_recorded_metadata: bool = True
+):
+    """
+    Produces a zipfile package of useful documentation supporting in silico
+    reproducibility of some QIIME 2 Result(s) from a QIIME 2 Artifact or
+    directory of Artifacts.
+
+    Package includes:
+    - replay scripts for all supported interfaces
+    - a bibtex-formatted collection of all citations
+    """
+    from qiime2.core.archive.provenance_lib.replay import replay_supplement
+    from qiime2.sdk.util import get_available_usage_drivers
+
+    usage_drivers = get_available_usage_drivers()
+    usage_driver_types = list(usage_drivers.values())
+    if not usage_driver_types:
+        msg = (
+            'There are no available usage drivers registered in the current '
+            'environment.'
+        )
+        raise ValueError(msg)
+
+    replay_supplement(
+        usage_drivers=usage_driver_types,
+        payload=in_fp,
+        out_fp=out_fp,
+        validate_checksums=validate_checksums,
+        parse_metadata=parse_metadata,
+        use_recorded_metadata=use_recorded_metadata,
+        recurse=recurse,
+        deduplicate=deduplicate,
+        suppress_header=suppress_header,
+        verbose=verbose,
+        dump_recorded_metadata=dump_recorded_metadata
+    )
