@@ -1,5 +1,5 @@
 # ----------------------------------------------------------------------------
-# Copyright (c) 2016-2022, QIIME 2 development team.
+# Copyright (c) 2016-2023, QIIME 2 development team.
 #
 # Distributed under the terms of the Modified BSD License.
 #
@@ -23,7 +23,12 @@ class GeneratedOption(click.Option):
         if metadata is not None:
             prefix = 'm'
         if multiple is not None:
-            multiple = list if multiple == 'list' else set
+            if multiple == 'list':
+                multiple = list
+            elif multiple == 'dict':
+                multiple = dict
+            else:
+                multiple = set
 
         if is_bool_flag:
             yes = q2cli.util.to_cli_name(name)
@@ -102,13 +107,22 @@ class GeneratedOption(click.Option):
 
     def _consume_metadata(self, ctx, opts):
         # double consume
+        # this consume deals with the metadata file
         md_file, source = super().consume_value(ctx, opts)
         # consume uses self.name, so mutate but backup for after
         backup, self.name = self.name, self.q2_extra_dest
-        md_col, _ = super().consume_value(ctx, opts)
+        try:
+            # this consume deals with the metadata column
+            md_col, _ = super().consume_value(ctx, opts)
+        # If `--m-metadata-column` isn't provided, need to set md_col to None
+        # in order for the click.MissingParameter errors below to be raised
+        except click.MissingParameter:
+            md_col = None
 
         self.name = backup
-
+        # These branches won't get hit unless there's a value associated with
+        # md_col - the try/except case above handled the situation where the
+        # metadata_column parameter itself wasn't provided (vs just a value)
         if (md_col is None) != (md_file is None):
             # missing one or the other
             if md_file is None:
@@ -180,16 +194,44 @@ class GeneratedOption(click.Option):
                 return None
             elif self.q2_prefix == 'i':
                 value = super().type_cast_value(ctx, value)
+                keys, value = self._split_and_validate_input_keys(value)
+
                 if self.q2_multiple is set:
                     self._check_length(value, ctx)
-                value = self.q2_multiple(value)
+
+                # This means we loaded a proper Collection directory. When we
+                # load in a Collection directory for an action that takes a
+                # Collection input, we get a tuple containing a dictionary of
+                # the Collection we wanted. When we load in a Collection
+                # directory for an action that takes a List, we get a list
+                # containing a dictionary of the Collection we wanted. We just
+                # extract that dictionary.
+                if (isinstance(value, tuple) or isinstance(value, list)) \
+                        and len(value) == 1 and isinstance(value[0], dict):
+                    value = value[0]
+
+                # We already have a dict, so we already have keys
+                if isinstance(value, dict):
+                    keys = value.keys()
+                    value = list(value.values())
+                elif self.q2_multiple is dict:
+                    if keys is None:
+                        keys = range(len(value))
+                    value = value
+                else:
+                    value = self.q2_multiple(value)
+
                 type_expr = qiime2.sdk.util.type_from_ast(self.q2_ast)
                 args = ', '.join(map(repr, (x.type for x in value)))
+
                 if value not in type_expr:
                     raise click.BadParameter(
                         'received <%s> as an argument, which is incompatible'
                         ' with parameter type: %r' % (args, type_expr),
                         ctx=ctx, param=self)
+
+                if self.q2_multiple is dict:
+                    value = {str(k): v for k, v in zip(keys, value)}
                 return value
             elif self.q2_metadata == 'file':
                 value = super().type_cast_value(ctx, value)
@@ -206,9 +248,43 @@ class GeneratedOption(click.Option):
                             e, header=header, traceback=tb)
             elif self.q2_prefix == 'p':
                 try:
+                    _values = []
+
                     if self.q2_multiple is set:
                         self._check_length(value, ctx)
-                    value = qiime2.sdk.util.parse_primitive(self.q2_ast, value)
+
+                    keys = []
+                    if self.q2_multiple is dict and type(value) is not dict:
+                        _values = {}
+
+                        keyed = False
+                        unkeyed = False
+                        # All params in a Collection must be either keyed or
+                        # unkeyed. We cannot have a mix because it makes things
+                        # ambiguous
+                        for idx, item in enumerate(value):
+                            if ':' in item:
+                                if unkeyed:
+                                    raise KeyError(
+                                        'The keyed value <%s> has been mixed'
+                                        ' with unkeyed values. All values must'
+                                        ' be keyed or unkeyed.' % item)
+                                key, _value = item.split(':', 1)
+                                _values[key] = _value
+                                keyed = True
+                            else:
+                                if keyed:
+                                    raise KeyError(
+                                        'The unkeyed value <%s> has been'
+                                        ' mixed with keyed values. All values'
+                                        ' must be keyed or unkeyed.' % item)
+                                _values[str(idx)] = item
+                                unkeyed = True
+                    else:
+                        _values = value
+
+                    value = \
+                        qiime2.sdk.util.parse_primitive(self.q2_ast, _values)
                 except ValueError:
                     args = ', '.join(map(repr, value))
                     expr = qiime2.sdk.util.type_from_ast(self.q2_ast)
@@ -217,13 +293,55 @@ class GeneratedOption(click.Option):
                         ' with parameter type: %r' % (args, expr),
                         ctx=ctx, param=self)
                 return value
+        elif self.q2_prefix == 'i':
+            value = super().type_cast_value(ctx, value)
+            if value is not None:
+                return value[1]
+            return value
 
+        # We have an output here
         return super().type_cast_value(ctx, value)
 
+    def _split_and_validate_input_keys(self, value):
+        """ This function ensures that if a user passed in a de-facto
+            collection they did so properly.
+        """
+        keys = [t[0] for t in value]
+        values = [t[1] for t in value]
+
+        if any(key is not None and not key.isidentifier() for key in keys):
+            raise ValueError('All keys must be valid Python identifiers.'
+                             ' Python identifier rules may be found here'
+                             ' https://www.askpython.com/python/'
+                             'python-identifiers-rules-best-practices')
+
+        # If we had no keys, we are fine
+        if all(key is None for key in keys):
+            return None, values
+
+        has_nones = any(key is None for key in keys)
+        has_keys = any(key is not None for key in keys)
+
+        # We cannot have keys for something that isn't a dict
+        if self.q2_multiple is not dict and has_keys:
+            raise ValueError('Keyed values may only be supplied for '
+                             'Collection inputs.')
+        # We cannot have a mixture of keyed and unkeyed values
+        elif self.q2_multiple is dict and has_keys and has_nones:
+            raise ValueError('Keyed values cannot be mixed with unkeyed '
+                             'values.')
+
+        return keys, values
+
     def _check_length(self, value, ctx):
         import collections
 
+        if isinstance(value, tuple) and len(value) == 1 and \
+                isinstance(value[0], dict):
+            value = list(value[0].values())
+
         counter = collections.Counter(value)
+
         dups = ', '.join(map(repr, (v for v, n in counter.items() if n > 1)))
         args = ', '.join(map(repr, value))
         if dups:
