@@ -1,5 +1,5 @@
 # ----------------------------------------------------------------------------
-# Copyright (c) 2016-2022, QIIME 2 development team.
+# Copyright (c) 2016-2023, QIIME 2 development team.
 #
 # Distributed under the terms of the Modified BSD License.
 #
@@ -8,6 +8,8 @@
 
 import click
 
+from qiime2.core.type.util import is_collection_type
+
 
 def is_writable_dir(path):
     import os
@@ -93,7 +95,9 @@ class QIIME2Type(click.ParamType):
                 self.fail('%r is already a directory.' % (value,), param, ctx)
 
         directory = os.path.dirname(value)
-        if directory and not os.path.exists(directory):
+
+        if (directory and not os.path.exists(directory)
+                and not is_collection_type(param.type.type_expr)):
             self.fail('Directory %r does not exist, cannot save %r into it.'
                       % (directory, os.path.basename(value)), param, ctx)
 
@@ -110,6 +114,8 @@ class QIIME2Type(click.ParamType):
 
         try:
             result, error = q2cli.util._load_input(value)
+            if result is not None:
+                result_value = result[1]
         except Exception as e:
             header = f'There was a problem loading {value!r} as an artifact:'
             q2cli.util.exit_with_error(
@@ -120,7 +126,7 @@ class QIIME2Type(click.ParamType):
         # We want to use click's fail to pretty print whatever error we got
         # from get_input
 
-        if isinstance(result, qiime2.sdk.Visualization):
+        if isinstance(result_value, qiime2.sdk.Visualization):
             maybe = value[:-1] + 'a'
             hint = ''
             if os.path.exists(maybe):
@@ -128,16 +134,15 @@ class QIIME2Type(click.ParamType):
                         ' %r, did you mean that?)'
                         % os.path.basename(maybe))
 
-            self.fail('%r is a QIIME 2 visualization (.qzv), not an '
+            self.fail('%r is a QIIME 2 visualization (.qzv), not an'
                       ' Artifact (.qza)%s' % (value, hint), param, ctx)
 
         style = qiime2.sdk.util.interrogate_collection_type(self.type_expr)
-        if style.style is None and result not in self.type_expr:
+        if style.style is None and result_value not in self.type_expr:
             # collections need to be handled above this
             self.fail("Expected an artifact of at least type %r."
                       " An artifact of type %r was provided."
-                      % (self.type_expr, result.type), param, ctx)
-
+                      % (self.type_expr, result_value.type), param, ctx)
         return result
 
     def _convert_metadata(self, value, param, ctx):
