@@ -1,5 +1,5 @@
 # ----------------------------------------------------------------------------
-# Copyright (c) 2016-2022, QIIME 2 development team.
+# Copyright (c) 2016-2023, QIIME 2 development team.
 #
 # Distributed under the terms of the Modified BSD License.
 #
@@ -8,6 +8,7 @@
 
 import os.path
 import unittest
+import contextlib
 import unittest.mock
 import tempfile
 import shutil
@@ -15,10 +16,10 @@ import click
 import errno
 
 from click.testing import CliRunner
-from qiime2 import Artifact, Visualization
 from qiime2.core.cache import get_cache
-from qiime2.core.testing.type import IntSequence1, IntSequence2
+from qiime2.core.testing.type import IntSequence1, IntSequence2, SingleInt
 from qiime2.core.testing.util import get_dummy_plugin
+from qiime2.sdk import Artifact, Visualization, ResultCollection
 
 from q2cli.builtin.info import info
 from q2cli.builtin.tools import tools
@@ -52,6 +53,7 @@ class CliTests(unittest.TestCase):
         self.assertIn('System versions', result.output)
         self.assertIn('Installed plugins', result.output)
         self.assertIn('dummy-plugin', result.output)
+        self.assertIn('other-plugin', result.output)
 
     def test_list_commands(self):
         # top level commands, including a plugin, are present
@@ -60,6 +62,7 @@ class CliTests(unittest.TestCase):
         self.assertIn('info', commands)
         self.assertIn('tools', commands)
         self.assertIn('dummy-plugin', commands)
+        self.assertIn('other-plugin', commands)
 
     def test_plugin_list_commands(self):
         # plugin commands are present including a method and visualizer
@@ -69,8 +72,26 @@ class CliTests(unittest.TestCase):
         self.assertIn('split-ints', commands)
         self.assertIn('mapping-viz', commands)
 
-        self.assertFalse('split_ints' in commands)
-        self.assertFalse('mapping_viz' in commands)
+        self.assertNotIn('split_ints', commands)
+        self.assertNotIn('mapping_viz', commands)
+        self.assertNotIn('-underscore-method', commands)
+        self.assertNotIn('_underscore-method', commands)
+
+    def test_plugin_list_hidden_commands(self):
+        # plugin commands are present including a method and visualizer and
+        # hidden method
+        qiime_cli = RootCommand()
+        command = qiime_cli.get_command(ctx=None, name='dummy-plugin')
+        commands = self.runner.invoke(command,
+                                      ['--show-hidden-actions']).output
+        self.assertIn('split-ints', commands)
+        self.assertIn('mapping-viz', commands)
+        self.assertIn('_underscore-method', commands)
+
+        self.assertNotIn('split_ints', commands)
+        self.assertNotIn('mapping_viz', commands)
+        self.assertNotIn('_underscore_method', commands)
+        self.assertNotIn('-underscore-method', commands)
 
     def test_action_parameter_types(self):
         qiime_cli = RootCommand()
@@ -93,29 +114,15 @@ class CliTests(unittest.TestCase):
             '10', '--output-dir', output_dir, '--verbose'])
         self.assertEqual(result.exit_code, 0)
 
-    def test_show_importable_types(self):
+    def test_execute_hidden_action(self):
+        int_path = os.path.join(self.tempdir, 'int.qza')
+        qiime_cli = RootCommand()
+        command = qiime_cli.get_command(ctx=None, name='dummy-plugin')
         result = self.runner.invoke(
-            tools, ['import', '--show-importable-types'])
-        self.assertEqual(result.exit_code, 0)
-        self.assertIn('FourInts', result.output)
-        self.assertIn('IntSequence1', result.output)
-        self.assertIn('IntSequence2', result.output)
-        self.assertIn('Kennel[Cat]', result.output)
-        self.assertIn('Kennel[Dog]', result.output)
-        self.assertIn('Mapping', result.output)
+                command, ['_underscore-method', '--o-int', int_path,
+                          '--verbose'])
 
-    def test_show_importable_formats(self):
-        result = self.runner.invoke(
-            tools, ['import', '--show-importable-formats'])
         self.assertEqual(result.exit_code, 0)
-        self.assertIn('FourIntsDirectoryFormat', result.output)
-        self.assertIn('IntSequenceDirectoryFormat', result.output)
-        self.assertNotIn('UnimportableFormat', result.output)
-        self.assertNotIn('UnimportableDirectoryFormat', result.output)
-        self.assertIn('MappingDirectoryFormat', result.output)
-        self.assertIn('IntSequenceFormat', result.output)
-        self.assertIn('IntSequenceFormatV2', result.output)
-        self.assertIn('IntSequenceV2DirectoryFormat', result.output)
 
     def test_extract(self):
         result = self.runner.invoke(
@@ -290,11 +297,6 @@ class CliTests(unittest.TestCase):
     def test_input_conversion(self):
         obj = QIIME2Type(IntSequence1.to_ast(), repr(IntSequence1))
 
-        with self.assertRaisesRegex(click.exceptions.BadParameter,
-                                    f'{self.tempdir!r} is a directory,'
-                                    ' not a QIIME 2 Artifact'):
-            obj._convert_input(self.tempdir, None, None)
-
         with self.assertRaisesRegex(click.exceptions.BadParameter,
                                     "x does not exist"):
             obj._convert_input('x', None, None)
@@ -567,7 +569,8 @@ class TestMetadataSupport(MetadataTestsBase):
 
 
 class TestMetadataColumnSupport(MetadataTestsBase):
-    def test_required_missing(self):
+    # Neither md file or column params provided
+    def test_required_missing_file_and_column(self):
         result = self._run_command(
             'identity-with-metadata-column', '--i-ints', self.input_artifact,
             '--o-out', self.output_artifact)
@@ -576,6 +579,27 @@ class TestMetadataColumnSupport(MetadataTestsBase):
         self.assertTrue(result.output.startswith('Usage:'))
         self.assertIn("Missing option '--m-metadata-file'", result.output)
 
+    # md file param missing, md column param & value provided
+    def test_required_missing_file(self):
+        result = self._run_command(
+            'identity-with-metadata-column', '--i-ints', self.input_artifact,
+            '--m-metadata-column', 'a', '--o-out', self.output_artifact)
+
+        self.assertEqual(result.exit_code, 1)
+        self.assertTrue(result.output.startswith('Usage:'))
+        self.assertIn("Missing option '--m-metadata-file'", result.output)
+
+    # md file param & value provided, md column param missing
+    def test_required_missing_column(self):
+        result = self._run_command(
+            'identity-with-metadata-column', '--i-ints', self.input_artifact,
+            '--m-metadata-file', self.metadata_file1,
+            '--o-out', self.output_artifact)
+
+        self.assertEqual(result.exit_code, 1)
+        self.assertTrue(result.output.startswith('Usage:'))
+        self.assertIn("Missing option '--m-metadata-column'", result.output)
+
     def test_optional_metadata_missing(self):
         result = self._run_command(
             'identity-with-optional-metadata-column', '--i-ints',
@@ -710,5 +734,279 @@ class TestMetadataColumnSupport(MetadataTestsBase):
         self.assertIn("expected Numeric", result.output)
 
 
+class TestCollectionSupport(unittest.TestCase):
+    def setUp(self):
+        get_dummy_plugin()
+        self.runner = CliRunner()
+        self.plugin_command = RootCommand().get_command(
+            ctx=None, name='dummy-plugin')
+        self.tempdir = tempfile.mkdtemp(prefix='qiime2-q2cli-test-temp-')
+
+        self.art1_path = os.path.join(self.tempdir, 'art1.qza')
+        self.art2_path = os.path.join(self.tempdir, 'art2.qza')
+        self.art1 = Artifact.import_data(SingleInt, 0)
+        self.art2 = Artifact.import_data(SingleInt, 1)
+
+        self.output = os.path.join(self.tempdir, 'out')
+        self.output2 = os.path.join(self.tempdir, 'out2')
+
+    def tearDown(self):
+        shutil.rmtree(self.tempdir)
+
+    def _run_command(self, *args):
+        return self.runner.invoke(self.plugin_command, args)
+
+    def test_collection_roundtrip_list(self):
+        result = self._run_command(
+            'list-params', '--p-ints', '0', '--p-ints', '1', '--o-output',
+            self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['0'].view(int), 0)
+        self.assertEqual(collection['1'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['0', '1'])
+
+        result = self._run_command(
+            'list-of-ints', '--i-ints', self.output, '--o-output',
+            self.output2, '--verbose'
+        )
+
+        self.assertEqual(collection['0'].view(int), 0)
+        self.assertEqual(collection['1'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['0', '1'])
+
+    def test_collection_roundtrip_dict_keyed(self):
+        result = self._run_command(
+            'dict-params', '--p-ints', 'foo:0', '--p-ints', 'bar:1',
+            '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['foo'].view(int), 0)
+        self.assertEqual(collection['bar'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['foo', 'bar'])
+
+        result = self._run_command(
+            'dict-of-ints', '--i-ints', self.output, '--o-output',
+            self.output2, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['foo'].view(int), 0)
+        self.assertEqual(collection['bar'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['foo', 'bar'])
+
+    def test_collection_roundtrip_dict_unkeyed(self):
+        result = self._run_command(
+            'dict-params', '--p-ints', '0', '--p-ints', '1',
+            '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['0'].view(int), 0)
+        self.assertEqual(collection['1'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['0', '1'])
+
+        result = self._run_command(
+            'dict-of-ints', '--i-ints', self.output, '--o-output',
+            self.output2, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['0'].view(int), 0)
+        self.assertEqual(collection['1'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['0', '1'])
+
+    def test_de_facto_list(self):
+        self.art1.save(self.art1_path)
+        self.art2.save(self.art2_path)
+
+        result = self._run_command(
+            'list-of-ints', '--i-ints', self.art1_path, '--i-ints',
+            self.art2_path, '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['0'].view(int), 0)
+        self.assertEqual(collection['1'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['0', '1'])
+
+    def test_de_facto_dict_keyed(self):
+        self.art1.save(self.art1_path)
+        self.art2.save(self.art2_path)
+
+        result = self._run_command(
+            'dict-of-ints', '--i-ints', f'foo:{self.art1_path}', '--i-ints',
+            f'bar:{self.art2_path}', '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['foo'].view(int), 0)
+        self.assertEqual(collection['bar'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['foo', 'bar'])
+
+    def test_de_facto_dict_unkeyed(self):
+        self.art1.save(self.art1_path)
+        self.art2.save(self.art2_path)
+
+        result = self._run_command(
+            'dict-of-ints', '--i-ints', self.art1_path, '--i-ints',
+            self.art2_path, '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['0'].view(int), 0)
+        self.assertEqual(collection['1'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['0', '1'])
+
+    def test_mixed_keyed_unkeyed_inputs(self):
+        self.art1.save(self.art1_path)
+        self.art2.save(self.art2_path)
+
+        result = self._run_command(
+            'dict-of-ints', '--i-ints', f'foo:{self.art1_path}', '--i-ints',
+            self.art2_path, '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 1)
+        self.assertIn('Keyed values cannot be mixed with unkeyed values.',
+                      str(result.exception))
+
+        result = self._run_command(
+            'dict-of-ints', '--i-ints', self.art1_path, '--i-ints',
+            f'bar:{self.art2_path}', '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 1)
+        self.assertIn('Keyed values cannot be mixed with unkeyed values.',
+                      str(result.exception))
+
+    def test_mixed_keyed_unkeyed_params(self):
+        result = self._run_command(
+            'dict-params', '--p-ints', 'foo:0', '--p-ints', '1',
+            '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 1)
+        self.assertIn('The unkeyed value <1> has been mixed with keyed values.'
+                      ' All values must be keyed or unkeyed',
+                      str(result.exception))
+
+        result = self._run_command(
+            'dict-params', '--p-ints', '0', '--p-ints', 'bar:1',
+            '--o-output', self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 1)
+        self.assertIn('The keyed value <bar:1> has been mixed with unkeyed'
+                      ' values. All values must be keyed or unkeyed',
+                      str(result.exception))
+
+    def test_keyed_path_with_tilde(self):
+        self.art1.save(self.art1_path)
+        self.art2.save(self.art2_path)
+
+        tmp = tempfile.gettempdir()
+        tempdir = os.path.basename(self.tempdir)
+
+        with modified_environ(HOME=tmp):
+            result = self._run_command(
+                'dict-of-ints',
+                '--i-ints', f'foo:{os.path.join("~", tempdir, "art1.qza")}',
+                '--i-ints', f'bar:{os.path.join("~", tempdir, "art2.qza")}',
+                '--o-output', self.output, '--verbose')
+
+        self.assertEqual(result.exit_code, 0)
+        collection = ResultCollection.load(self.output)
+
+        self.assertEqual(collection['foo'].view(int), 0)
+        self.assertEqual(collection['bar'].view(int), 1)
+        self.assertEqual(list(collection.keys()), ['foo', 'bar'])
+
+    def test_directory_with_non_artifacts(self):
+        input_dir = os.path.join(self.tempdir, 'in')
+        os.mkdir(input_dir)
+
+        artifact_path = os.path.join(input_dir, 'a.qza')
+        artifact = Artifact.import_data(IntSequence1, [0, 42, 43])
+        artifact.save(artifact_path)
+
+        with open(os.path.join(input_dir, 'bad.txt'), 'w') as fh:
+            fh.write('This file is not an artifact')
+
+        result = self._run_command(
+            'list-of-ints', '--i-ints', input_dir, '--o-output',
+            self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 1)
+        self.assertIn("Invalid value for '--i-ints':", result.output)
+
+    def test_empty_directory(self):
+        result = self._run_command(
+            'list-of-ints', '--i-ints', self.tempdir, '--o-output',
+            self.output, '--verbose'
+        )
+
+        self.assertEqual(result.exit_code, 1)
+        self.assertIn(f"Provided directory '{self.tempdir}' is empty.",
+                      result.output)
+
+
+@contextlib.contextmanager
+def modified_environ(*remove, **update):
+    """
+    Taken from: https://stackoverflow.com/a/34333710.
+
+    Updating the os.environ dict only modifies the environment variables from
+    the perspective of the current Python process, so this isn't dangerous at
+    all.
+
+    Temporarily updates the ``os.environ`` dictionary in-place.
+
+    The ``os.environ`` dictionary is updated in-place so that the modification
+    is sure to work in all situations.
+
+    :param remove: Environment variables to remove.
+    :param update: Dictionary of environment variables and values to
+                   add/update.
+    """
+    env = os.environ
+    update = update or {}
+    remove = remove or []
+
+    # List of environment variables being updated or removed.
+    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
+    # Environment variables and values to restore on exit.
+    update_after = {k: env[k] for k in stomped}
+    # Environment variables and values to remove on exit.
+    remove_after = frozenset(k for k in update if k not in env)
+
+    try:
+        env.update(update)
+        [env.pop(k, None) for k in remove]
+        yield
+    finally:
+        env.update(update_after)
+        [env.pop(k) for k in remove_after]
+
+
 if __name__ == "__main__":
     unittest.main()
